#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "resource.h"
#include "resource\BGFont.h"
//#include "resource\BitMapFont.h"

#include <stdlib.h>

//---------------------------------------------------------------------------
//Xe[W̃TCY
#define STAGE_CX            15
#define STAGE_CY            9

//Xe[WɎgubN
#define STAGE_BLANK         0
#define STAGE_GREEN         1
#define STAGE_BLUE          2
#define STAGE_YELLOW        3
#define STAGE_PINK          4
#define STAGE_BROWN         5
#define STAGE_TYPE_COUNT    5	//BLANK͐ɂȂł

//ZBG^C
#define BG_TILE_BLACK       0
#define BG_TILE_GREEN       2
#define BG_TILE_BLUE        4
#define BG_TILE_YELLOW      6
#define BG_TILE_PINK        8
#define BG_TILE_BROWN       10

//Q[̐is
#define GAME_STATE_TITLE    1	//Q[̃^Cg\
#define GAME_STATE_INIT     2	//Xe[W̏
#define GAME_STATE_INPUT    3	//
#define GAME_STATE_DELETE   4	//͈ʒũubN폜
#define GAME_STATE_FILL     5	//폜ubNl߂
#define GAME_STATE_ENDCHK   6	//Q[̏I

//̓L[̃EFCg
#define KEY_RESPONSE        1

#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)>(b))?(b):(a))
//---------------------------------------------------------------------------
//J[\̒`
typedef struct {
	s16 x;
	s16 y;
	u16 cx;
	u16 cy;
	s16 mx;				//ړ
	s16 my;	
	s16 MinX;			//ړ
	s16 MinY;
	s16 MaxX;
	s16 MaxY;
	u16 SprNo;			//XvCg
	u16 SprTile;
	u16 SprSize;
	u16 SprForm;
	u8  key;			//ړp
} ST_CURSOR;

ST_CURSOR CURSOR;

//---------------------------------------------------------------------------
u8 BG2Chr;
u8 BG2Map;
u8 BG2Col;

u8 BG3Chr;
u8 BG3Map;
u8 BG3Col;

u8   Stage [STAGE_CX][STAGE_CY];
bool Select[STAGE_CX][STAGE_CY];	//ubN̏mFp

u16 Score;

u8  GameStateFlag;		//Q[̐is
u16 KeyCount;			//L[EFCgp
u16 KeyOld;				//L[EFCgp
//---------------------------------------------------------------------------
void Init();
void InitCursor();
void InitStage();
void InitSelect();

void GameStateTitle();
void GameStateInit();
void GameStateInput();
void GameStateDelete();
void GameStateFill();
void GameStateEndChk();

void DrawStageAll();
void DrawStage(const u8 x, const u8 y, const u8 type);
void DrawBGStage(const u8 bgX, const u8 bgY, const u8 tileNo);
void DrawBGScore();
void DrawBGTitle();
void DrawBGGameOver();
void DrawSprCursor();

void Cursor();
void MoveCursor(ST_CURSOR* p);
u8   ChkMoveCursor(ST_CURSOR* p);

void SelectBlock(s8 x, s8 y, u8 mark);
u8   GetSelectCount();

void DeleteBlock();

void FillDownBlock();
void FillLeftBlock();

bool IsGameOver();

void CalcScore(u8 selectCnt);
void CalcBouns();
//---------------------------------------------------------------------------
int main(void)
{
	Init();
	InitCursor();
	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	for(;;)
	{
		WaitForVsync();

		switch(GameStateFlag)
		{
			case GAME_STATE_TITLE:  GameStateTitle();  break;
			case GAME_STATE_INIT:   GameStateInit();   break;
			case GAME_STATE_INPUT:  GameStateInput();  break;
			case GAME_STATE_DELETE: GameStateDelete(); break;
			case GAME_STATE_FILL:   GameStateFill();   break;
			case GAME_STATE_ENDCHK: GameStateEndChk(); break;
			default: DebugMessage((u8*)"Err: Main GameStateFlag"); break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void Init()
{
	//EFCgRg[
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;

	//L[
	KeyInit();
	KeyCount = 0;
	KeyOld   = 0;

	//BG
	//BG2 \p
	//BG3 Q[p
	BG2Chr = 1;
	BG2Map = 0;
	BG2Col = 0;

	BG3Chr = 2;
	BG3Map = 1;
	BG3Col = 1;

	REG_BG2CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG2Chr) | BG_MAPBASE(BG2Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG3Chr) | BG_MAPBASE(BG3Map) | 0);

	BGSetPal16  (BG2Col, (u16*)&BGFont_palette);
	BGSetChrData(BG2Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3Col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3Chr, (u16*)&game_bg_img_bin, 2048);
	BGSetPalMap (BG3Map, 2048, 1);

	//Sprite
	SpriteInit();
	SpriteSetPal16((u16*)&game_spr_pal_bin);
	SpriteSetData ((u16*)&game_spr_img_bin, 2048);

	//Q[̐ݒ
	GameStateFlag = GAME_STATE_TITLE;
}
//---------------------------------------------------------------------------
void InitCursor()
{
	CURSOR.x       = SCREEN_CX;
	CURSOR.y       = SCREEN_CY;
	CURSOR.cx      = 16;
	CURSOR.cy      = 16;
	CURSOR.mx      = CURSOR.cx;		//ړʂ̐ݒ
	CURSOR.my      = CURSOR.cy;
	CURSOR.MinX    = 0;				//ړ̐ݒ
	CURSOR.MaxX    = SCREEN_CX;
	CURSOR.MinY    = 0;
	CURSOR.MaxY    = SCREEN_CY - CURSOR.cy;
	CURSOR.key     = 0x00;

	CURSOR.SprNo   = 0;
	CURSOR.SprTile = 0;
	CURSOR.SprSize = SP_SIZE_16;
	CURSOR.SprForm = SP_SQUARE;

	SpriteSetSize(CURSOR.SprNo, CURSOR.SprSize, CURSOR.SprForm, SP_COLOR_16);
	SpriteSetChr (CURSOR.SprNo, CURSOR.SprTile);
	SpriteMove   (CURSOR.SprNo, CURSOR.x, CURSOR.y);
}
//---------------------------------------------------------------------------
void GameStateTitle()
{
	DrawBGTitle();
	WaitKey(KEY_A);		//srand
	BGDrawCls(BG2Map, 0x00);

	GameStateFlag = GAME_STATE_INIT;
}
//---------------------------------------------------------------------------
void GameStateInit()
{
	InitStage();
	InitSelect();

	DrawStageAll();
	DrawBGScore();

	CURSOR.x = 0;
	CURSOR.y = 0;
	DrawSprCursor();

	GameStateFlag = GAME_STATE_INPUT;
}
//---------------------------------------------------------------------------
void DrawStageAll()
{
	u8 x,y;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			DrawStage(x, y, Stage[x][y]);
		}
	}
}
//---------------------------------------------------------------------------
void DrawStage(const u8 x, const u8 y, const u8 type)
{
	u8 bgX    = x*2;	//Xe[WWafW֕ϊ
	u8 bgY    = y*2;
	u8 tileNo = 0x00;

	switch(type)
	{
		case STAGE_BLANK:  tileNo = BG_TILE_BLACK;  break;
		case STAGE_GREEN:  tileNo = BG_TILE_GREEN;  break;
		case STAGE_BLUE:   tileNo = BG_TILE_BLUE;   break;
		case STAGE_YELLOW: tileNo = BG_TILE_YELLOW; break;
		case STAGE_PINK:   tileNo = BG_TILE_PINK;   break;
		case STAGE_BROWN:  tileNo = BG_TILE_BROWN;  break;
		default: DebugMessage((u8*)"Err: DrawStage type"); break;
	}

	DrawBGStage(bgX, bgY, tileNo);
}
//---------------------------------------------------------------------------
void DrawBGStage(const u8 bgX, const u8 bgY, const u8 tileNo)
{
	BGDrawChar(BG3Map, bgX+0, bgY+0, BG3Col, tileNo+0 );
	BGDrawChar(BG3Map, bgX+1, bgY+0, BG3Col, tileNo+1 );
	BGDrawChar(BG3Map, bgX+0, bgY+1, BG3Col, tileNo+32);
	BGDrawChar(BG3Map, bgX+1, bgY+1, BG3Col, tileNo+33);
}
//---------------------------------------------------------------------------
void GameStateInput()
{
	Cursor();
}
//---------------------------------------------------------------------------
void Cursor()
{
	u8 Key = KeyInput();

	//L[X|X(TODO ꃉCuj
	if(Key == 0x00 && KeyOld != Key)
	{
		KeyCount = 0;
		KeyOld   = Key;
		return;
	}
	if(KeyCount++ <= KEY_RESPONSE)
	{
		return;
	}

	KeyCount = 0;
	CURSOR.key = 0x00;

	if(Key & KEY_UP   ) CURSOR.key |= KEY_UP;
	if(Key & KEY_DOWN ) CURSOR.key |= KEY_DOWN;
	if(Key & KEY_LEFT ) CURSOR.key |= KEY_LEFT;
	if(Key & KEY_RIGHT) CURSOR.key |= KEY_RIGHT;
	if(Key & KEY_A    ) CURSOR.key  = KEY_A;
	if(Key & KEY_B    ) CURSOR.key  = KEY_B;

	//ړ`FbN
	if( ChkMoveCursor(&CURSOR) == TRUE )
	{
		MoveCursor(&CURSOR);
		DrawSprCursor();
	}
	//ubN̍폜
	if(Key & KEY_A)
	{
		GameStateFlag = GAME_STATE_DELETE;

		//ςȂ֎~܂
		WaitKey(KEY_A);
	}
	if(Key & KEY_B)
	{
		GameStateFlag = GAME_STATE_INIT;
	}
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
u8 ChkMoveCursor(ST_CURSOR* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->key & KEY_UP   ) ty1-=p->my;
	if(p->key & KEY_DOWN ) ty1+=p->my;
	if(p->key & KEY_LEFT ) tx1-=p->mx;
	if(p->key & KEY_RIGHT) tx1+=p->mx;

	if(tx1< CURSOR.MinX || ty1< CURSOR.MinY) return FALSE;
	if(tx1>=CURSOR.MaxX || ty1>=CURSOR.MaxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
//ړ
void MoveCursor(ST_CURSOR* p)
{
	if(p->key & KEY_UP   ) p->y-=p->my;
	if(p->key & KEY_DOWN ) p->y+=p->my;
	if(p->key & KEY_LEFT ) p->x-=p->mx;
	if(p->key & KEY_RIGHT) p->x+=p->mx;
}
//---------------------------------------------------------------------------
void DrawSprCursor()
{
	SpriteMove(CURSOR.SprNo, CURSOR.x, CURSOR.y);
}
//---------------------------------------------------------------------------
void DrawBGScore()
{
	char msg[40];
	bprintf(msg, "%5d", Score);
	BGDrawStr(BG2Map, 25, 19, (u8*)msg);
}
//---------------------------------------------------------------------------
void DrawBGTitle()
{
	BGDrawStr(BG2Map, 1, 2, (u8*)"Push A Button");
}
//---------------------------------------------------------------------------
void DrawBGGameOver()
{
	BGDrawStr(BG2Map, 1, 19, (u8*)"Game Over");
}
//---------------------------------------------------------------------------
void GameStateDelete()
{
	s8 curX = CURSOR.x / CURSOR.cx;
	s8 curY = CURSOR.y / CURSOR.cy;
	u8 mark = Stage[curX][curY];

	//\ȃubN܂
	InitSelect();
	SelectBlock(curX, curY, mark);

	//\ȃubN̐JEg܂
	u8 selectCnt = GetSelectCount();
	if(selectCnt <= 1)
	{
		GameStateFlag = GAME_STATE_INPUT;
		return;
	}

	DeleteBlock();

	CalcScore(selectCnt);
	DrawBGScore();

	GameStateFlag = GAME_STATE_FILL;
}
//---------------------------------------------------------------------------
void GameStateFill()
{
	u8 cnt;
	for(cnt=0; cnt<STAGE_CY; cnt++)
	{
		FillDownBlock();
	}
	for(cnt=0; cnt<STAGE_CX; cnt++)
	{
		FillLeftBlock();
	}

	DrawStageAll();
	GameStateFlag = GAME_STATE_ENDCHK;
}
//---------------------------------------------------------------------------
//\ȃubN܂
void SelectBlock(s8 x, s8 y, u8 mark)
{
	if(x < 0 || x > STAGE_CX-1) return;
	if(y < 0 || y > STAGE_CY-1) return;

	if(Stage[x][y] != mark || Stage[x][y] == STAGE_BLANK)
	{
		return;
	}
	if(Select[x][y] == TRUE)
	{
		return;
	}

	Select[x][y] = TRUE;

	SelectBlock(x+0,y+1,mark);
	SelectBlock(x+0,y-1,mark);
	SelectBlock(x+1,y+0,mark);
	SelectBlock(x-1,y+0,mark);
}
//---------------------------------------------------------------------------
u8 GetSelectCount()
{
	u8 x,y;
	u8 retCnt = 0;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			if(Select[x][y] == TRUE)
			{
				retCnt++;
			}
		}
	}

	return retCnt;
}
//---------------------------------------------------------------------------
void DeleteBlock()
{
	u8 x,y;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			if(Select[x][y] == TRUE)
			{
				Stage[x][y] = STAGE_BLANK;
			}
		}
	}
}
//---------------------------------------------------------------------------
void CalcScore(u8 selectCnt)
{
	if(selectCnt < 2)
	{
		return;
	}

	Score += (selectCnt-2) * (selectCnt-2);
}
//---------------------------------------------------------------------------
void InitStage()
{
	u8 x,y;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			Stage [x][y] = rand() % STAGE_TYPE_COUNT + 1;
		}
	}
}
//---------------------------------------------------------------------------
void InitSelect()
{
	u8 x,y;

	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			Select[x][y] = FALSE;
		}
	}
}
//---------------------------------------------------------------------------
//ubNɋl߂܂
void FillDownBlock()
{
	u8 x, y, y2;

	for(x=0; x<STAGE_CX; x++)
	{
		for(y=0; y<STAGE_CY; y++)
		{
			if(Stage[x][y] == STAGE_BLANK)
			{
				//󔒂l߂ăRs[܂
				for(y2=y; y2>=1; y2--)
				{
					Stage[x][y2  ] = Stage[x][y2-1];
					Stage[x][y2-1] = STAGE_BLANK;
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
//ubNɋl߂܂
void FillLeftBlock()
{
	bool BlankFlag;
	u8   x, y, x2, y2;

	for(x=0; x<STAGE_CX; x++)
	{
		BlankFlag = TRUE;

		//c̗񂪂ׂċ󔒂`FbN܂
		for(y=0; y<STAGE_CY; y++)
		{
			if(Stage[x][y] != STAGE_BLANK)
			{
				BlankFlag = FALSE;
			}
		}

		//c̗񂪂ׂċ󔒂ꍇAɋl߂܂
		if(BlankFlag == TRUE)
		{
			for(x2=x; x2<STAGE_CX-1; x2++)
			{
				for(y2=0; y2<STAGE_CY; y2++)
				{
					Stage[x2  ][y2] = Stage[x2+1][y2];
					Stage[x2+1][y2] = STAGE_BLANK;
				}
			}
		}
	}
}
//---------------------------------------------------------------------------
void GameStateEndChk()
{
	if(IsGameOver() == TRUE)
	{
		CalcBouns();
		DrawBGScore();
		DrawBGGameOver();

		WaitKey(KEY_B);
		BGDrawCls(BG2Map, 0x00);

		GameStateFlag = GAME_STATE_INIT;
	}
	else
	{
		GameStateFlag = GAME_STATE_INPUT;
	}
}
//---------------------------------------------------------------------------
//{[iX̌vZ܂
void CalcBouns()
{
	u8 x,line=0;

	for(x=0; x<STAGE_CX; x++)
	{
		if(Stage[x][STAGE_CY-1] != STAGE_BLANK)
		{
			line++;
		}
	}

	if(line == 0)
	{
		Score *= 3;
	}
	else if(line < 10)
	{
		Score *= 2.0 - 0.1*line;
	}
}
//---------------------------------------------------------------------------
bool IsGameOver()
{
	u8 x, y;
	u8 selectCnt;

	for(x=0; x<STAGE_CX; x++)
	{
		for(y=0; y<STAGE_CY; y++)
		{
			InitSelect();
			SelectBlock(x, y, Stage[x][y]);

			selectCnt = GetSelectCount();

			if(selectCnt >= 2)
			{
				return FALSE;
			}
		}
	}
	return TRUE;
}
